/*****************************************************************************/
/*
                                 Watch.c

The WATCH facility is a powerful adjunct in server administration.  From the
administration menu it provides an online, real-time, in-browser-window view of
request processing in the running server.   The ability to observe live request
processing on an ad hoc basis, without changing server configuration or
shutting-down/restarting the server process, makes this facility a great
configuration and problem resolution tool.  It allows (amongst other uses)

  o  assessment of mapping rules
  o  assessment of authorization rules
  o  investigation of request processing problems
  o  general observation of server behaviour

A single client per server process can access the WATCH facility at any one
time. The report can be generated for a user-specified number of seconds or
aborted at any time using the browser's stop button.

An event is considered any significant point for which the server code has a
reporting call provided. These have been selected to provide maximum
information with minimum clutter and impact on server performance. Obvious
examples are connection acceptance and closure, request path resolution, error
report generation, network reads and writes, etc.

This module also provides functions to display a process using SHOW PROCESS
/ALL via the DCL.C module, and to delete a process, as well as a function to
"peek" at selected fields in the data structures of any request during
processing (intended more as a diagnosis and development tool).


VERSION HISTORY
---------------
10-NOV-1999  MGD  use decc$match_wild() for the WatchFilter()
30-OCT-1999  MGD  dignified with a module of it's own (unbundled from ADMIN.C)
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application related header files */
#include "wasd.h"

#define WASD_MODULE "WATCH"

/******************/
/* global storage */
/******************/

#define WATCH_FILTER_SIZE 128

int  WatchDisabled,
     WatchEnabled,
     WatchOneShotEnabled,
     WatchItem;

char  WatchClientFilter [WATCH_FILTER_SIZE],
      WatchPathFilter [WATCH_FILTER_SIZE],
      WatchServiceFilter [WATCH_FILTER_SIZE];

struct RequestStruct  *WatchRequestPtr;

char  ErrorWatchAuthNeeded [] =
         "Authorization must be enabled to access WATCH!",
      ErrorWatchBufferOvf [] = "*ERROR* sys$fao() BUFFEROVF",
      ErrorWatchNumber [] = "Not found in current request list.",
      ErrorWatchSelf [] = "Cannot WATCH yourself!",
      ErrorWatchQueryString [] = "WATCH query string problem.",
      ErrorWatchSysFao [] = "*ERROR* sys$fao()";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern boolean  LoggingEnabled,
                ProtocolHttpsConfigured,
                ProxyCacheEnabled,
                ProxyServingEnabled;

extern unsigned long HttpdProcessId;

extern char  BuildInfo[],
             CommandLine[],
             ErrorSanityCheck[],
             HttpdUserName[],
             ServerHostPort[],
             SoftwareID[],
             SyiHwName[],
             SyiVersion[],
             TimeGmtString[],
             Utility[];

extern struct MsgStruct  Msgs;
extern struct ListHeadStruct  RequestList;

/*****************************************************************************/
/*
Return a multi-line string of package identification information including some
quick-and-dirty image analysis.  Open the UCX$IPC_SHR shareable image and read
selected fields to get the image name, identification, and linking date.  Don't
quite know how to get this information using header structures, etc., too busy
(that's my excuse) to investigate ... hence this quick mangle.  Also try to
establish which package is involved by checking for likely logical names.
*/ 

char* WatchPackageInfo (struct RequestStruct *rqptr)

{
   static char  ErrorInfo [32],
                ImageInfoDateString [32];
   static char  *PackageInfoPtr = NULL;

   static $DESCRIPTOR (PackageFaoDsc,
"!AZ (!AZ)\n\
!AZ-TCP/IP !AC !AC (!AZ)\n\
!AZ\
!AZ VMS !AZ\n\
$ HTTPD !AZ\n");

   static $DESCRIPTOR (BufferDsc, "");
   static $DESCRIPTOR (DateFaoDsc, "!%D\0");
   static $DESCRIPTOR (ImageInfoDateStringDsc, ImageInfoDateString);
   static $DESCRIPTOR (LnmSystemDsc, "LNM$SYSTEM");
   static $DESCRIPTOR (MultinetDsc, "MULTINET");
   static $DESCRIPTOR (TcpWareDsc, "TCPWARE");
   static $DESCRIPTOR (TcpIpExamplesDsc, "TCPIP$EXAMPLES");
   static $DESCRIPTOR (UcxExamplesDsc, "UCX$EXAMPLES");

   int  status;
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [16];
   char  *PackagePtr,
         *ImageIdentPtr,
         *ImageNamePtr;
   char  Buffer [1024],
         ImageInfo [512];
   FILE  *ImageFile;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchPackageInfo()\n");

   if (PackageInfoPtr != NULL) return (PackageInfoPtr);

   if ((ImageFile =
       fopen ("UCX$IPC_SHR", "r", "shr=get", "dna=SYS$LIBRARY:.EXE"))
       == NULL)
   {
      sprintf (ErrorInfo+1, "%%X%08.08X", vaxc$errno);
      ErrorInfo[0] = strlen(ErrorInfo+1);
      ImageNamePtr = ImageIdentPtr = ErrorInfo;
      memset (&ImageInfo, 0, sizeof(ImageInfo));
   }
   else
   if (!fread (&ImageInfo, sizeof(ImageInfo), 1, ImageFile))
   {
      sprintf (ErrorInfo+1, "%%X%08.08X", vaxc$errno);
      ErrorInfo[0] = strlen(ErrorInfo+1);
      ImageNamePtr = ImageIdentPtr = ErrorInfo;
      memset (&ImageInfo, 0, sizeof(ImageInfo));
   }
   else
   {
#ifdef __ALPHA
      ImageNamePtr = &ImageInfo[0xc8];
      ImageIdentPtr = &ImageInfo[0xf0];
      sys$fao (&DateFaoDsc, 0, &ImageInfoDateStringDsc,
               &ImageInfo[0xc0]);
#else
      Length = *(unsigned short*)ImageInfo;
      ImageNamePtr = &ImageInfo[Length-0x50];
      ImageIdentPtr = &ImageInfo[Length-0x28];
      sys$fao (&DateFaoDsc, 0, &ImageInfoDateStringDsc,
               &ImageInfo[Length-0x18]);
#endif
      if (ImageInfoDateString[0] == ' ') ImageInfoDateString[0] = '0';
   }

   fclose (ImageFile);

   PackagePtr = "TCPware";
   status = sys$trnlnm (0, &LnmSystemDsc, &TcpWareDsc, 0, 0);
   if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   if (status == SS$_NOLOGNAM)
   {
      PackagePtr = "Multinet";
      status = sys$trnlnm (0, &LnmSystemDsc, &MultinetDsc, 0, 0);
      if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   }
   if (status == SS$_NOLOGNAM)
   {
      /* Compaq TCP/IP (UC) v5.n-> */
      PackagePtr = "Compaq";
      status = sys$trnlnm (0, &LnmSystemDsc, &TcpIpExamplesDsc, 0, 0);
      if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   }
   if (status == SS$_NOLOGNAM)
   {
      PackagePtr = "UCX";
      status = sys$trnlnm (0, &LnmSystemDsc, &UcxExamplesDsc, 0, 0);
      if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   }
   if (status == SS$_NOLOGNAM)
      PackagePtr = "unknown";
   else
   if (VMSnok (status))
      PackagePtr = "sys$trnlnm()_error";

   vecptr = &FaoVector;
   *vecptr++ = SoftwareID;
   *vecptr++ = BuildInfo;
   *vecptr++ = PackagePtr;
   *vecptr++ = ImageNamePtr;
   *vecptr++ = ImageIdentPtr;
   *vecptr++ = ImageInfoDateString;
   *vecptr++ = SesolaVersion();
   *vecptr++ = SyiHwName;
   *vecptr++ = SyiVersion;
   *vecptr++ = CommandLine;

   BufferDsc.dsc$a_pointer = Buffer;
   BufferDsc.dsc$w_length = sizeof(Buffer)-1;

   sys$faol (&PackageFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return ("");
   }
   Buffer[Length] = '\0';

   PackageInfoPtr = VmGet(Length+1);
   strcpy (PackageInfoPtr, Buffer);

   return (PackageInfoPtr);
}                                   

/*****************************************************************************/
/*
Check if the watch facility is already in use, report error if it is.  Provide
form for selection of WATCH parameters.
*/ 

WatchReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... WATCH Report</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>WATCH Report</H2>\n\
!AZ\n\
\
<FORM ACTION=\"!AZ\">\n\
<P><TABLE BORDER=1 CELLSPACING=0 CELLPADDING=5>\n\
<TR><TH>Select WATCH Criteria</TH></TR>\n\
<TR><TD>\n\
<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=5>\n\
<TR><TD VALIGN=top><NOBR>\n\
<B><U>Request</U></B>\n\
<BR><INPUT TYPE=checkbox NAME=rqp VALUE=1 CHECKED> Processing\n\
<BR><INPUT TYPE=checkbox NAME=rqh VALUE=1> Header\n\
<BR><INPUT TYPE=checkbox NAME=rqb VALUE=1> Body\n\
<BR><B><U>Response</U></B>\n\
<BR><INPUT TYPE=checkbox NAME=rsp VALUE=1 CHECKED> Processing\n\
<BR><INPUT TYPE=checkbox NAME=rsh VALUE=1> Header\n\
<BR><INPUT TYPE=checkbox NAME=rsb VALUE=1> Body\n\
</NOBR></TD><TD VALIGN=top><NOBR>\n\
<B><U>General</U></B>\n\
<BR><INPUT TYPE=checkbox NAME=con VALUE=1 CHECKED> Connection\n\
<BR><INPUT TYPE=checkbox NAME=map VALUE=1> Mapping\n\
<BR><INPUT TYPE=checkbox NAME=aut VALUE=1> Authorization\n\
<BR><INPUT TYPE=checkbox NAME=err VALUE=1> Error\n\
<BR><INPUT TYPE=checkbox NAME=cgi VALUE=1> CGI\n\
<BR><INPUT TYPE=checkbox NAME=dcl VALUE=1> DCL\n\
<BR><INPUT TYPE=checkbox NAME=dnt VALUE=1> DECnet\n\
</NOBR></TD><TD VALIGN=top><NOBR>\n\
<B><U>Network</U></B>\n\
<BR><INPUT TYPE=checkbox NAME=net VALUE=1> Activity\n\
<BR><INPUT TYPE=checkbox NAME=oct VALUE=1> Data\n\
<BR><B><U>Other</U></B>\n\
<BR><INPUT TYPE=checkbox NAME=tmr VALUE=1> Timer\n\
<BR><INPUT TYPE=checkbox NAME=log VALUE=1> !AZLogging!AZ\n\
<BR><INPUT TYPE=checkbox NAME=ssl VALUE=1> !AZSSL!AZ\n\
</NOBR></TD><TD VALIGN=top><NOBR>\n\
<B><U>Proxy</U></B>\n\
!AZ\
<BR><INPUT TYPE=checkbox NAME=pxy VALUE=1> Processing\n\
<BR><INPUT TYPE=checkbox NAME=prh VALUE=1> Request Header\n\
<BR><INPUT TYPE=checkbox NAME=prb VALUE=1> Request Body\n\
<BR><INPUT TYPE=checkbox NAME=psh VALUE=1> Response Header\n\
<BR><INPUT TYPE=checkbox NAME=psb VALUE=1> Response Body\n\
!AZ\
!AZ\
<BR><INPUT TYPE=checkbox NAME=pca VALUE=1> Cache\n\
<BR><INPUT TYPE=checkbox NAME=pcm VALUE=1> Cache Maintenance\n\
!AZ\
</NOBR></TD></TR>\n\
<TR><TD COLSPAN=4>\n\
<INPUT TYPE=text NAME=clf SIZE=40 VALUE=\"*\"> Client Filter\n\
<BR><INPUT TYPE=text NAME=sef SIZE=40 VALUE=\"*\"> Service Filter\n\
<BR><INPUT TYPE=text NAME=paf SIZE=40 VALUE=\"*\"> Path Filter\n\
<BR><SELECT NAME=dul>\n\
<OPTION> 30\n\
<OPTION SELECTED> 60\n\
<OPTION> 120\n\
<OPTION> 300\n\
<OPTION> 600\n\
</SELECT> or\n\
<INPUT TYPE=text NAME=dut SIZE=5 VALUE=\"\"> Seconds Duration\n\
<P><INPUT TYPE=submit VALUE=\" WATCH \">\n\
<INPUT TYPE=reset VALUE=\" reset \">\n\
</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</FORM>\n\
\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [16];
   char  Buffer [4096];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchReport()\n");

   if (WatchDisabled)
   {
      rqptr->ResponseStatusCode = 501;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_DISABLED), FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (!rqptr->RemoteUser[0])
   {
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, ErrorWatchAuthNeeded, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (WatchInUse (rqptr))
   {
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);
   *vecptr++ = INTERNAL_REPORT_WATCH;
   if (LoggingEnabled)
   {
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "<I>";
      *vecptr++ = "</I>";
   }
   if (ProtocolHttpsConfigured)
   {
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "<I>";
      *vecptr++ = "</I>";
   }
   if (ProxyServingEnabled)
   {
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "<I>\n";
      *vecptr++ = "</I>\n";
   }
   if (ProxyCacheEnabled)
   {
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = "<I>\n";
      *vecptr++ = "</I>\n";
   }

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, NextTaskFunction, Buffer, Length);
}

/*****************************************************************************/
/*
Check if the WATCH facility is already in use, report error if it is.  Parse
the query string WATCH parameters.  Report any errors.  Place the parameters
into the WATCH global storage.  This reserves the WATCH facility for this
client via the 'WatchRequestPtr' (and indicate this with a flag in the request
structure, which will be detected as the request concludes and the WATCH
facility released for reuse).  Generate a plain-text HTTP header and output a
WATCH report heading.  Then return BUT do not declare any AST to continue
processing.  The client will just "hang" there receiving output from
WatchThis() via the structure pointed to by 'WatchRequestPtr'.  Can be used to
WATCH all new requests (matching any filter criteria of course) or in a
"one-shot" mode where a single request is selected to display all categories at
any point during it's processing.
*/ 

WatchBegin
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!20%D  WATCH REPORT  !AZ\n\
!#*-\n\
!AZ\
Watching: !AZ (!UL)\n\
Client: \"!AZ\" Service: \"!AZ\" Path: \"!AZ\"\n\
|Time_______|Module__|Line|Item|Category|Event...|\n");
   static char  DummyReadBuffer [32];

   register char  *cptr, *qptr, *sptr, *zptr;
   register struct RequestStruct  *rqeptr;
   register struct ListEntryStruct  *leptr;

   int  status,
        ConnectNumber,
        DurationSeconds,
        WatchCategories;
   char  Buffer [2048],
         CategoryList [512],
         ClientFilter [WATCH_FILTER_SIZE],
         FieldName [256],
         FieldValue [WATCH_FILTER_SIZE],
         PathFilter [WATCH_FILTER_SIZE],
         ServiceFilter [WATCH_FILTER_SIZE];
   unsigned short  Length;
   unsigned long  Category;
   unsigned long  *vecptr;
   unsigned long  FaoVector [16];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchBegin()\n");

   if (WatchDisabled)
   {
      rqptr->ResponseStatusCode = 501;
      ErrorGeneral (rqptr, MsgFor(rqptr,MSG_GENERAL_DISABLED), FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (!rqptr->RemoteUser[0])
   {
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, ErrorWatchAuthNeeded, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (WatchInUse (rqptr))
   {
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   ConnectNumber = DurationSeconds = WatchCategories = 0;
   ClientFilter[0] = PathFilter[0] = ServiceFilter[0] = '\0';
   if (rqptr->QueryStringLength)
   {
      qptr = rqptr->QueryStringPtr;
      while (*qptr)
      {
         qptr = ParseQueryField (rqptr, qptr,
                                 FieldName, sizeof(FieldName),
                                 FieldValue, sizeof(FieldValue),
                                 FI_LI);
         if (strsame (FieldName, "one", -1) && FieldValue[0])
         {
            WatchCategories = WATCH_ONE_SHOT;
            ConnectNumber = strtol (FieldValue, NULL, 10);
            if (Debug) fprintf (stdout, "ConnectNumber: %d\n", ConnectNumber);
         }
         else
         if (strsame (FieldName, "aut", -1) && FieldValue[0])
            WatchCategories |= WATCH_AUTH;
         else
         if (strsame (FieldName, "cgi", -1) && FieldValue[0])
            WatchCategories |= WATCH_CGI;
         else
         if (strsame (FieldName, "con", -1) && FieldValue[0])
            WatchCategories |= WATCH_CONNECT;
         else
         if (strsame (FieldName, "dcl", -1) && FieldValue[0])
            WatchCategories |= WATCH_DCL;
         else
         if (strsame (FieldName, "dnt", -1) && FieldValue[0])
            WatchCategories |= WATCH_DECNET;
         else
         if (strsame (FieldName, "err", -1) && FieldValue[0])
            WatchCategories |= WATCH_ERROR;
         else
         if (strsame (FieldName, "log", -1) && FieldValue[0])
            WatchCategories |= WATCH_LOG;
         else
         if (strsame (FieldName, "map", -1) && FieldValue[0])
            WatchCategories |= WATCH_MAPPING;
         else
         if (strsame (FieldName, "net", -1) && FieldValue[0])
            WatchCategories |= WATCH_NETWORK;
         else
         if (strsame (FieldName, "oct", -1) && FieldValue[0])
            WatchCategories |= WATCH_NETWORK_OCTETS;
         else
         if (strsame (FieldName, "pxy", -1) && FieldValue[0])
            WatchCategories |= WATCH_PROXY;
         else
         if (strsame (FieldName, "pca", -1) && FieldValue[0])
            WatchCategories |= WATCH_PROXY_CACHE;
         else
         if (strsame (FieldName, "pcm", -1) && FieldValue[0])
            WatchCategories |= WATCH_PROXY_CACHE_MNT;
         else
         if (strsame (FieldName, "prh", -1) && FieldValue[0])
            WatchCategories |= WATCH_PROXY_REQU_HDR;
         else
         if (strsame (FieldName, "prb", -1) && FieldValue[0])
            WatchCategories |= WATCH_PROXY_REQU_BDY;
         else
         if (strsame (FieldName, "psh", -1) && FieldValue[0])
            WatchCategories |= WATCH_PROXY_RESP_HDR;
         else
         if (strsame (FieldName, "psb", -1) && FieldValue[0])
            WatchCategories |= WATCH_PROXY_RESP_BDY;
         else
         if (strsame (FieldName, "rqp", -1) && FieldValue[0])
            WatchCategories |= WATCH_REQUEST;
         else
         if (strsame (FieldName, "rqb", -1) && FieldValue[0])
            WatchCategories |= WATCH_REQUEST_BODY;
         else
         if (strsame (FieldName, "rqh", -1) && FieldValue[0])
            WatchCategories |= WATCH_REQUEST_HEADER;
         else
         if (strsame (FieldName, "rsp", -1) && FieldValue[0])
            WatchCategories |= WATCH_RESPONSE;
         else
         if (strsame (FieldName, "rsb", -1) && FieldValue[0])
            WatchCategories |= WATCH_RESPONSE_BODY;
         else
         if (strsame (FieldName, "rsh", -1) && FieldValue[0])
            WatchCategories |= WATCH_RESPONSE_HEADER;
         else
         if (strsame (FieldName, "ssl", -1) && FieldValue[0])
            WatchCategories |= WATCH_SESOLA;
         else
         if (strsame (FieldName, "tmr", -1) && FieldValue[0])
            WatchCategories |= WATCH_TIMER;
         else
         if (strsame (FieldName, "clf", -1))
            strcpy (ClientFilter, FieldValue);
         else
         if (strsame (FieldName, "paf", -1))
            strcpy (PathFilter, FieldValue);
         else
         if (strsame (FieldName, "sef", -1))
            strcpy (ServiceFilter, FieldValue);
         else
         if (strsame (FieldName, "dul", -1) ||
             strsame (FieldName, "dut", -1) ||
             strsame (FieldName, "sec", -1) ||
             strsame (FieldName, "seconds", -1))
         {
            for (cptr = FieldValue; *cptr && !isdigit(*cptr); cptr++);
            if (*cptr) DurationSeconds = atoi (cptr);
         }
         else
         {
            rqptr->ResponseStatusCode = 403;
            ErrorGeneral (rqptr, ErrorWatchQueryString, FI_LI);
            SysDclAst (NextTaskFunction, rqptr);
            return;
         }
      }
   }

   if (WatchCategories == WATCH_ONE_SHOT && !DurationSeconds)
      DurationSeconds = WATCH_ONE_SHOT_DEFAULT_SECONDS;

   if (!WatchCategories || !DurationSeconds)
   {
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, ErrorWatchQueryString, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   if (ConnectNumber)
   {
      /* find this connection number in the current request list */
      for (leptr = RequestList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         rqeptr = (struct RequestStruct*)leptr;
         if (Debug)
            fprintf (stdout, "ConnectNumber: %d\n", rqeptr->ConnectNumber);
         if (rqeptr->ConnectNumber == ConnectNumber) break;
      }
      if (leptr == NULL)
      {
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, ErrorWatchNumber, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      if (rqeptr == rqptr)
      {
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, ErrorWatchSelf, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      rqeptr->WatchItem = WATCH_ONE_SHOT_ITEM;
   }

   if (!ClientFilter[0]) strcpy (ClientFilter, "*");
   if (!PathFilter[0]) strcpy (PathFilter, "*");
   if (!ServiceFilter[0]) strcpy (ServiceFilter, "*");

   zptr = (sptr = CategoryList) + 80;
   for (Category = 1; Category; Category = Category << 1)
   {
      if ((cptr = WatchCategory (WatchCategories & Category)) != NULL)
      {
         if (sptr > CategoryList)
         {
            *sptr++ = ',';
            if (sptr > zptr)
            {
               zptr = sptr + 80;
               *sptr++ = '\n';
            }
            else
               *sptr++ = ' ';
         }
         while (*cptr) *sptr++ = tolower(*cptr++);
      }
   }
   *sptr = '\0';

   if (Debug)
      fprintf (stdout, "%08.08X |%s| %d |%s|%s|%s|\n",
               WatchCategories, CategoryList, DurationSeconds,
               ClientFilter, PathFilter, ServiceFilter);

   /* now we grab the WATCH facility */
   WatchRequestPtr = rqptr;
   WatchEnabled = WatchCategories;
   WatchItem = 0;
   strcpy (WatchClientFilter, ClientFilter);
   strcpy (WatchPathFilter, PathFilter);
   strcpy (WatchServiceFilter, ServiceFilter);

   /* make sure we get the duration we asked for! */
   rqptr->TimerOutputCount =
      rqptr->TimerNoProgressCount =
      rqptr->TimerNoProgressSeconds = DurationSeconds;

   rqptr->ResponsePreExpired = PRE_EXPIRE_WATCH;
   HTTP_HEADER_200_PLAIN (rqptr);

   vecptr = FaoVector;
   *vecptr++ = 0;
   *vecptr++ = ServerHostPort;
   *vecptr++ = 36 + strlen(ServerHostPort);
   *vecptr++ = WatchPackageInfo(rqptr);
   *vecptr++ = CategoryList;
   *vecptr++ = WatchCategories;
   *vecptr++ = WatchClientFilter;
   *vecptr++ = WatchServiceFilter;
   *vecptr++ = WatchPathFilter;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   /* turn a leading date space into a leading zero */
   if (Buffer[0] == ' ' ) Buffer[0] = '0';

   NetWrite (rqptr, 0, Buffer, Length);

   NetRead (rqptr, &WatchDummyReadAst,
            DummyReadBuffer, sizeof(DummyReadBuffer));

   /* the request now just "hangs", reading WATCH plain-text output! */
}

/*****************************************************************************/
/*
Check and report if the WATCH facility is already being used.
*/ 

boolean WatchInUse (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (InUseFaoDsc, "WATCH is currently in use by !AZ!AZ!AZ");

   int  status;
   char  Buffer [256];
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [8];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchInUse()\n");

   if (WatchRequestPtr == NULL && !WatchEnabled) return (false);

   vecptr = FaoVector;
   if (WatchRequestPtr == NULL)
   {
      *vecptr++ = "/WATCH";
      *vecptr++ = "";
      *vecptr++ = "";
   }
   else
   {
      *vecptr++ = WatchRequestPtr->RemoteUser;
      *vecptr++ = "@";
      *vecptr++ = WatchRequestPtr->ServicePtr->ClientHostName;
   }
   status = sys$faol (&InUseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      return (true);
   }
   Buffer[Length] = '\0';
   rqptr->ResponseStatusCode = 403;
   ErrorGeneral (rqptr, Buffer, FI_LI);
   return (true);
}                                   

/*****************************************************************************/
/*
Request using the WATCH facility drops the connection.  Release WATCH.
*/ 

WatchEnd (struct RequestStruct *rqptr)

{

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchEnd()\n");

   if (WatchRequestPtr != rqptr)
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);

   WatchRequestPtr = NULL;
   WatchEnabled = WatchItem = 0;
   WatchClientFilter[0] =
      WatchPathFilter[0] =
      WatchServiceFilter[0] = '\0';
}                                   

/*****************************************************************************/
/*
The WATCH client concluding the watching can only be detected via a break in
connection which in a quiescent system (no requests being processed) can in
turn only be detected by a broken network read I/O.  This dummy read ASTs to
here where the status can be checked and appropriate action taken.
*/ 

WatchDummyReadAst (struct RequestStruct *rqptr)

{
   static char  DummyReadBuffer [32];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
         "WatchDummyReadAst() NetReadIOsb.Status %%X%08.08X .Count %d\n",
         rqptr->NetReadIOsb.Status, rqptr->NetReadIOsb.Count);

   if (VMSnok (rqptr->NetReadIOsb.Status))
   {
      RequestEnd (rqptr);
      return;
   }

   /* just in case it wasn't an error! */
   NetRead (rqptr, &WatchDummyReadAst,
            DummyReadBuffer, sizeof(DummyReadBuffer));
}                                   

/*****************************************************************************/
/*
Filter requests WATCHed on the basis of originating client and/or request
contents (service connected, request path).  Return true if the request should
be WATCHed, false if not.
*/ 

boolean WatchFilter
(
char *ClientHostName,
char *ClientHostAddress,
char *RequestSchemeName,
char *ServerHostPort,
char *PathInfo
)
{
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "WatchFilter() |%s|%s|%s|\n",
               WatchClientFilter, WatchServiceFilter,
               WatchPathFilter);

   /* this can occur if ASTs are delivered after WATCH use is discontinued */
   if (!WatchEnabled) return (false);

   if (ClientHostName != NULL || ClientHostAddress != NULL)
   {
      /* filter on client host name/address */
      if (*(unsigned short*)WatchClientFilter != '*\0')
      {
         if (isdigit(*(sptr = WatchClientFilter)))
            cptr = ClientHostAddress;
         else
            cptr = ClientHostName;
         if (Debug) fprintf (stdout, "cptr |%s|\n", cptr);
         if (cptr == NULL) return (true);
         if (!decc$match_wild (cptr, sptr)) return (false);
      }
   }

   if (ServerHostPort != NULL)
   {
      /* filter on service */
      if (Debug)
         fprintf (stdout, "ServerHostPort |%s|%s|\n",
                  RequestSchemeName, ServerHostPort);
      if (*(unsigned short*)WatchServiceFilter != '*\0')
      {
         sptr = WatchServiceFilter;
         if (!memcmp (sptr, "http:", 5))
         {
            if (strcmp (RequestSchemeName, "http:")) return (false);
            sptr += 5;
            while (*sptr == '/') sptr++;
         }
         else
         if (!memcmp (sptr, "https:", 6))
         {
            if (strcmp (RequestSchemeName, "https:")) return (false);
            sptr += 6;
            while (*sptr == '/') sptr++;
         }
         cptr = ServerHostPort;
         if (!decc$match_wild (cptr, sptr)) return (false);
      }
   }

   if (PathInfo != NULL)
   {
      /* filter on path */
      if (Debug) fprintf (stdout, "PathInfo |%s|\n", PathInfo);
      if (*(unsigned short*)WatchPathFilter != '*\0')
      {
         cptr = PathInfo;
         sptr = WatchPathFilter;
         if (!decc$match_wild (cptr, sptr)) return (false);
      }
   }

   return (true);
}                                   

/*****************************************************************************/
/*
VARIABLE LENGTH ARGUMENT LIST. Function to provide a formatted WATCH entry,
with trailing information from the caller. 'ReportFormat' parameter must be in
a sys$fao() acceptable format and sufficient variable number parameters be
supplied to satisfy any FAO directives in that format string. The report line
is then generated and output to either 'stdout' (the server process log if
/WATCH= qualifier was used) or to the request structure pointed to by the
gloabl storage 'WatchRequestPtr' (if a request-based WATCH is in use).
*/ 

WatchThis
(
struct RequestStruct *rqptr,
char *SourceModuleName,
int SourceLineNumber,
int Category,
char *ReportFormat,
...
)
{

   static $DESCRIPTOR (BufferFaoDsc, "|!%T !8AZ !4ZL !4ZL !8AZ !AZ|\n");
   static $DESCRIPTOR (ReportFaoDsc, "");

   register char  *cptr, *sptr;

   int  status,
        argcnt;
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [32];
   char  Buffer [2048],
         CategoryString [128],
         Report [2048];
   va_list  argptr;
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (ReportDsc, Report);

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug)
      fprintf (stdout, "WatchThis() %s %d |%s| %d\n",
               SourceModuleName, SourceLineNumber, ReportFormat, argcnt);

   /* this can occur if ASTs are delivered after WATCH use is discontinued */
   if (!WatchEnabled) return;

   vecptr = FaoVector;
   va_start (argptr, ReportFormat);
   for (argcnt -= 4; argcnt; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   ReportFaoDsc.dsc$a_pointer = ReportFormat;
   ReportFaoDsc.dsc$w_length = strlen(ReportFormat);

   status = sys$faol (&ReportFaoDsc, &Length, &ReportDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      if (Debug) fprintf (stdout, "sys$faol() %%X%08.08X\n", status);
      if (status == SS$_BUFFEROVF)
         strcpy (Report, ErrorWatchBufferOvf);
      else
         strcpy (Report, ErrorWatchSysFao);
   }
   else
      Report[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Report);

   if ((cptr = WatchCategory (Category)) == NULL)
      cptr = "????????";
   sptr = CategoryString;
   while (*cptr && *cptr != ' ') *sptr++ = *cptr++;
   *sptr = '\0';

   vecptr = FaoVector;
   *vecptr++ = 0;
   *vecptr++ = SourceModuleName;
   *vecptr++ = SourceLineNumber;
   if (rqptr == NULL)
      *vecptr++ = -1;
   else
   if (!rqptr->WatchItem)
      *vecptr++ = -1;
   else
      *vecptr++ = rqptr->WatchItem % 10000;
   *vecptr++ = CategoryString;
   *vecptr++ = Report;

   status = sys$faol (&BufferFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      if (Debug) fprintf (stdout, "sys$faol() %%X%08.08X\n", status);
      if (status == SS$_BUFFEROVF)
         strcpy (Report, ErrorWatchBufferOvf);
      else
         strcpy (Report, ErrorWatchSysFao);
   }
   else
      Buffer[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);

   if (WatchRequestPtr == NULL)
      fputs (Buffer, stdout);
   else
      NetWrite (WatchRequestPtr, 0, Buffer, Length);
}

/*****************************************************************************/
/*
VARIABLE LENGTH ARGUMENT LIST.  Function to provide a formatted data WATCH entry,
with trailing information from the caller.  'DataFormat' parameter must be
in a sys$fao() acceptable format and sufficient variable number parameters be
supplied to satisfy any FAO directives in that format string.  Should include
appropriate carriage-control.
*/ 

WatchDataFormatted
(
char *DataFormat,
...
)
{

   static $DESCRIPTOR (DataFaoDsc, "");

   int  status,
        argcnt;
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [32];
   char  Buffer [2048];
   va_list  argptr;
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug)
      fprintf (stdout, "WatchDataFormatted() |%s| %d\n",
               DataFormat, argcnt);

   /* this can occur if ASTs are delivered after WATCH use is discontinued */
   if (!WatchEnabled) return;

   vecptr = FaoVector;
   va_start (argptr, DataFormat);
   for (argcnt -= 1; argcnt; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   DataFaoDsc.dsc$a_pointer = DataFormat;
   DataFaoDsc.dsc$w_length = strlen(DataFormat);

   status = sys$faol (&DataFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      if (Debug) fprintf (stdout, "sys$faol() %%X%08.08X\n", status);
      if (status == SS$_BUFFEROVF)
         strcpy (Buffer, ErrorWatchBufferOvf);
      else
         strcpy (Buffer, ErrorWatchSysFao);
   }
   else
      Buffer[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);

   if (WatchRequestPtr == NULL)
      fputs (Buffer, stdout);
   else
      NetWrite (WatchRequestPtr, 0, Buffer, Length);
}

/*****************************************************************************/
/*
Output the supplied, non-formatted data in the WATCH report.  Allows any
printable output to be included as a block in the WATCH output.
*/ 

WatchData
(
char *DataPtr,
int DataLength
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchData()\n");

   /* this can occur if ASTs are delivered after WATCH use is discontinued */
   if (!WatchEnabled) return;

   if (WatchRequestPtr == NULL)
      fputs (DataPtr, stdout);
   else
      NetWrite (WatchRequestPtr, 0, DataPtr, DataLength);

   if (DataLength && DataPtr[DataLength-1] == '\n') return;

   if (WatchRequestPtr == NULL)
      fputs ("\n", stdout);
   else
      NetWrite (WatchRequestPtr, 0, "\n", 1);
}

/*****************************************************************************/
/*
Ouput the supplied data using WATCH as a hex and printable character dump.
*/ 

WatchDataDump
(
char *DataPtr,
int DataLength
)
{
/* 32 bytes by 128 lines comes out to 4096 bytes, the default buffer-full */
#define MAX_LINES 128
#define BYTES_PER_LINE 32
#define BYTES_PER_GROUP 4
#define GROUPS_PER_LINE (BYTES_PER_LINE / BYTES_PER_GROUP)
#define CHARS_PER_LINE ((BYTES_PER_LINE * 3) + GROUPS_PER_LINE + 1)

   static char  HexDigits [] = "0123456789ABCDEF";

   register char  *cptr, *sptr, *zptr;

   int  ByteCount,
        CurrentDataCount,
        DataCount;
   char  *CurrentDataPtr,
         *CurrentDumpPtr;
   char  DumpBuffer [(CHARS_PER_LINE * MAX_LINES)+1];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchDataDump() %d\n", DataLength);

   /* this can occur if ASTs are delivered after WATCH use is discontinued */
   if (!WatchEnabled) return;

   zptr = (sptr = DumpBuffer) + sizeof(DumpBuffer)-1;
   cptr = DataPtr;
   DataCount = DataLength;

   while (DataCount)
   {
      CurrentDumpPtr = sptr;
      CurrentDataPtr = cptr;
      CurrentDataCount = DataCount;

      ByteCount = BYTES_PER_LINE;
      while (ByteCount && DataCount)
      {
         *sptr++ = HexDigits[*(unsigned char*)cptr >> 4];
         *sptr++ = HexDigits[*(unsigned char*)cptr & 0xf];
         cptr++;
         DataCount--;
         ByteCount--;
         if (!(ByteCount % BYTES_PER_GROUP)) *sptr++ = ' ';
      }
      while (ByteCount)
      {
         *sptr++ = ' ';
         *sptr++ = ' ';
         ByteCount--;
         if (!(ByteCount % BYTES_PER_GROUP)) *sptr++ = ' ';
      }

      cptr = CurrentDataPtr;
      DataCount = CurrentDataCount;

      ByteCount = BYTES_PER_LINE;
      while (ByteCount && DataCount)
      {
         if (isalnum(*cptr) || ispunct(*cptr) || *cptr == ' ')
            *sptr++ = *cptr++;
         else
         {
            *sptr++ = '.';
            cptr++;
         }
         DataCount--;
         ByteCount--;
      }
      *sptr++ = '\n';

      if (Debug)
      {
         *sptr = '\0';
         /** fprintf (stdout, "|%s|\n", CurrentDumpPtr); **/
      }

      if (!DataCount || sptr >= zptr)
      {
         *sptr = '\0';
         WatchData (DumpBuffer, sptr - DumpBuffer);
         zptr = (sptr = DumpBuffer) + sizeof(DumpBuffer)-1;
      }
   }
}

/*****************************************************************************/
/*
*/ 

char* WatchCategory (int Category)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchCategory() %d\n", Category);

   switch (Category)
   {
      case WATCH_AUTH             : return ("AUTHORIZE");
      case WATCH_CONNECT          : return ("CONNECT");
      case WATCH_CGI              : return ("CGI");
      case WATCH_DCL              : return ("DCL");
      case WATCH_DECNET           : return ("DECNET");
      case WATCH_ERROR            : return ("ERROR");
      case WATCH_FILTER           : return ("FILTER");
      case WATCH_LOG              : return ("LOG");
      case WATCH_MAPPING          : return ("MAPPING");
      case WATCH_NETWORK          : return ("NETWORK");
      case WATCH_NETWORK_OCTETS   : return ("NETWORK-OCTETS");
      case WATCH_PROXY            : return ("PROXY");
      case WATCH_PROXY_CACHE      : return ("PROXY-CACHE");
      case WATCH_PROXY_CACHE_MNT  : return ("PROXY-CACHE-MAINTENANCE");
      case WATCH_PROXY_REQU_HDR   : return ("PROXY-REQUEST-HEADER");
      case WATCH_PROXY_REQU_BDY   : return ("PROXY-REQUEST-BODY");
      case WATCH_PROXY_RESP_HDR   : return ("PROXY-RESPONSE-HEADER");
      case WATCH_PROXY_RESP_BDY   : return ("PROXY-RESPONSE-BODY");
      case WATCH_REQUEST          : return ("REQUEST");
      case WATCH_REQUEST_BODY     : return ("REQUEST-BODY");
      case WATCH_REQUEST_HEADER   : return ("REQUEST-HEADER");
      case WATCH_RESPONSE         : return ("RESPONSE");
      case WATCH_RESPONSE_BODY    : return ("RESPONSE-BODY");
      case WATCH_RESPONSE_HEADER  : return ("RESPONSE-HEADER");
      case WATCH_SESOLA           : return ("SSL");
      case WATCH_TIMER            : return ("TIMER");
      default : return (NULL);
   }
}

/*****************************************************************************/
/*
Parse the /WATCH= qualifier string for command-line startup control.
*/ 

boolean WatchCli (char *String)

{
   register char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchCli() |%s|\n", String);

   if (strsame (String, "/NOWATCH", 6))
   {
      WatchDisabled = -1;
      return (true);
   }

   cptr = String;
   while (*cptr && *cptr != '=') cptr++;
   while (*cptr == '=' || *cptr == '\"') cptr++;
   if (!*cptr) return (true);
   if (Debug) fprintf (stdout, "cptr |%s|\n", cptr);
   if (!(WatchEnabled = atoi(cptr)))
   {
      fprintf (stdout, "%%%s-E-WATCH, invalid category number", Utility);
      return (false);
   }
   while (*cptr && (*cptr == '-' || isdigit(*cptr))) cptr++;
   if (*cptr == ',') cptr++;
   if (Debug) fprintf (stdout, "cptr |%s|\n", cptr);
   zptr = (sptr = WatchClientFilter) + sizeof(WatchClientFilter);
   while (*cptr && *cptr != ',' && *cptr != '\"' && sptr < zptr)
      *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      fprintf (stdout, "%%%s-E-WATCH, invalid client filter", Utility);
      return (false);
   }
   *sptr = '\0';
   if (*cptr == ',') cptr++;
   if (Debug) fprintf (stdout, "cptr |%s|\n", cptr);
   zptr = (sptr = WatchServiceFilter) + sizeof(WatchServiceFilter);
   while (*cptr && *cptr != ',' && *cptr != '\"' && sptr < zptr)
      *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      fprintf (stdout, "%%%s-E-WATCH, invalid service filter", Utility);
      return (false);
   }
   *sptr = '\0';
   if (*cptr == ',') cptr++;
   if (Debug) fprintf (stdout, "cptr |%s|\n", cptr);
   zptr = (sptr = WatchPathFilter) + sizeof(WatchPathFilter);
   while (*cptr && *cptr != ',' && *cptr != '\"' && sptr < zptr)
      *sptr++ = *cptr++;
   if (sptr >= zptr)
   {
      fprintf (stdout, "%%%s-E-WATCH, invalid path filter", Utility);
      return (false);
   }
   *sptr = '\0';

   if (!WatchClientFilter[0]) strcpy (WatchClientFilter, "*");
   if (!WatchPathFilter[0]) strcpy (WatchPathFilter, "*");
   if (!WatchServiceFilter[0]) strcpy (WatchServiceFilter, "*");

   return (true);
}

/*****************************************************************************/
/*
Just print out the WATCH CLI settings.
*/ 

WatchCliSettings ()

{
   register char  *cptr, *sptr;

   unsigned long  Category;
   char  CategoryList [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchCliSettings()\n");

   if (WatchDisabled)
      fprintf (stdout, "%%%s-I-WATCH, disabled\n", Utility);

   if (!WatchEnabled) return;

   sptr = CategoryList;
   for (Category = 1; Category; Category = Category << 1)
   {
      if ((cptr = WatchCategory (WatchEnabled & Category)) != NULL)
      {
         if (sptr > CategoryList)
         {
            *sptr++ = ',';
            *sptr++ = ' ';
         }
         while (*cptr) *sptr++ = tolower(*cptr++);
      }
   }
   *sptr = '\0';

   fprintf (stdout,
"%%%s-I-WATCH, (%d) %s\n\
-%s-I-WATCH, client filter \"%s\"\n\
-%s-I-WATCH, service filter \"%s\"\n\
-%s-I-WATCH, path filter \"%s\"\n",
            Utility, WatchEnabled, CategoryList,
            Utility, WatchClientFilter,
            Utility, WatchServiceFilter,
            Utility, WatchPathFilter);
}

/*****************************************************************************/
/*
Generate a report page listing all of the processes belonging to the server
process.
*/

#define PSCAN$_GETJPI_BUFFER_SIZE 24

WatchProcessReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Process Report</TITLE>\n\
</HEAD>\n\
<BODY LINK=\"#0000cc\" VLINK=\"#0000cc\">\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Process Report</H2>\n\
!AZ\n\
\
<P><TABLE BORDER=1 CELLSPACING=0 CELLPADDING=5>\n\
<TR><TH>!AZ</TH></TR>\n\
<TR><TD>\n\
<TABLE BORDER=0 CELLSPACING=0 CELLPADDING=5>\n\
<TR><TH></TH><TH ALIGN=left><U>PID<U></TH>\
<TH ALIGN=left><U>Process Name</U>&nbsp;</TH>\
<TH ALIGN=left><U>Image</U></TH>\
<TH ALIGN=left><U>State</U></TH></TR>\n");

   static $DESCRIPTOR (ProcessFaoDsc,
"<TR><TH ALIGN=right>!UL.&nbsp;</TH>\
<TD><A HREF=\"!AZ?pid=!8XL\">!8XL</A>&nbsp;</TD>\
<TD>!AZ&nbsp;</TD><TD>!AZ&nbsp;</TD><TD>!AZ</TD></TR>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   static char  *StateNameArray [] = {
                "1234","COLPG","MWAIT","CEF","PFW","LEF","LEFO",
                "HIB","HIBO","SUSP","SUSPO","FPG","COM","COMO","CUR" };

   static unsigned long  GetJpiControlFlags = JPI$M_IGNORE_TARGET_STATUS;

   static unsigned long  JpiPid,
                         JpiState;
   static char  JpiImagName [256],
                JpiNodeName [32],
                JpiPrcNam [16],
                JpiUserName [13];

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   }
      JpiItems [] =
   {
      { sizeof(GetJpiControlFlags), JPI$_GETJPI_CONTROL_FLAGS,
        &GetJpiControlFlags, 0 },
      { sizeof(JpiPid), JPI$_PID, &JpiPid, 0 },
      { sizeof(JpiImagName), JPI$_IMAGNAME, &JpiImagName, 0 },
      { sizeof(JpiPrcNam), JPI$_PRCNAM, &JpiPrcNam, 0 },
      { sizeof(JpiState), JPI$_STATE, &JpiState, 0 },
      { 0,0,0,0 }
   },
      ScanItems [] =
   {
      { 0, PSCAN$_GETJPI_BUFFER_SIZE, 2048, 0},
      { 0,0,0,0 }
   };

   register char  *cptr;
   register unsigned long  *vecptr;

   int  status,
        ProcessCount;
   unsigned short  Length;
   unsigned long  ProcessContext;
   unsigned long  FaoVector [16];
   char  *StateNamePtr;
   char  Buffer [4096];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchProcessReport()\n");

   ProcessContext = 0;
   status = sys$process_scan (&ProcessContext, &ScanItems);
   if (Debug) fprintf (stdout, "sys$process_scan() %%X%08.08X\n", status);
   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = "sys$getjpi()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);
   *vecptr++ = HttpdUserName;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   ProcessCount = 0;

   for (;;)
   {
      status = sys$getjpiw (0, &ProcessContext, 0, &JpiItems, 0, 0, 0);
      if (Debug) fprintf (stdout, "sys$getjpi() %%X%08.08X\n", status);
      if (VMSnok (status)) break;

      ProcessCount++;

      JpiUserName[12] = '\0';
      for (cptr = JpiUserName; *cptr && *cptr != ' '; cptr++);
      *cptr = '\0';
      if (Debug) fprintf (stdout, "JpiUserName |%s|\n", JpiUserName);

      JpiPrcNam[15] = '\0';
      for (cptr = JpiPrcNam; *cptr && *cptr != ' '; cptr++);
      *cptr = '\0';
      if (Debug) fprintf (stdout, "JpiPrcNam |%s|\n", JpiPrcNam);

      for (cptr = JpiImagName; *cptr && *cptr != ';'; cptr++);
      if (*cptr == ';') *cptr-- = '\0';
      while (cptr > JpiImagName && *cptr != ']') cptr--;
      if (*cptr == ']') cptr++;

      if (JpiState > 0 && JpiState <= 14)
         StateNamePtr = StateNameArray[JpiState];
      else
         sprintf (StateNamePtr = StateNameArray[0], "%04.04X", JpiState);

      vecptr = FaoVector;
      *vecptr++ = ProcessCount;
      *vecptr++ = INTERNAL_REPORT_SHOW_PROCESS;
      *vecptr++ = JpiPid;
      *vecptr++ = JpiPid;
      *vecptr++ = JpiPrcNam;
      if (*cptr)
         *vecptr++ = cptr;
      else
         *vecptr++ = "[<FONT SIZE=-1>DCL</FONT>]";
      *vecptr++ = StateNamePtr;

      status = sys$faol (&ProcessFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   if (status != SS$_NOMOREPROC)
   {
      rqptr->ErrorTextPtr = "sys$getjpi()";
      ErrorVmsStatus (rqptr, status, FI_LI);
   }

   NetWriteBuffered (rqptr, NextTaskFunction,
                     EndPageFaoDsc.dsc$a_pointer,
                     EndPageFaoDsc.dsc$w_length);
}

/*****************************************************************************/
/*
Using a scripting subprocess do a SHOW PROCESS /ALL on the specified process. 
Used from the DclReport() but actually could be used on any process the server 
has access to, including the server!
*/

WatchShowProcess
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *ProcessIdString
)
{
   static char  DclCommand [512];
   static unsigned long  JpiServerPid;
   static $DESCRIPTOR (DclCommandDsc, DclCommand);
   static $DESCRIPTOR (DclCommandFaoDsc,
"SHOW PROCESS /ALL /IDENT=!AZ\n\
SV=$SEVERITY\n\
IF SV THEN WS=\"WRITE SYS$OUTPUT\"\n\
IF SV THEN WS \"\"\n\
IF SV THEN WS \"Mode: \" + F$GETJPI(\"!AZ\",\"MODE\")\n\
IF SV THEN WS \"\"\n\
IF SV THEN IF F$GETJPI(\"!AZ\",\"IMAGNAME\") .EQS. \"\" \
THEN WS \"Image: [DCL]\"\n\
IF SV THEN IF F$GETJPI(\"!AZ\",\"IMAGNAME\") .NES. \"\" \
THEN WS \"Image: \" + F$GETJPI(\"!AZ\",\"IMAGNAME\")\0");

   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Show Process</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>Show Process</H2>\n\
!AZ\n\
\
<P><TABLE CELLSPACING=0 CELLPADDING=5 BORDER=1>\n\
<TR><TH>PID &nbsp;!AZ</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLSPACING=0 CELLPADDING=5 BORDER=0>\n\
<TR><TD><PRE>");

   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  ProcessId;
   unsigned long  FaoVector [32];
   char  Buffer [1024];
   void  *EndPageFunctionPtr;
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchShowProcess() %s\n", ProcessIdString);

   ProcessId = strtol (ProcessIdString, NULL, 16);
   if (Debug) fprintf (stdout, "ProcessId: %X08.08X\n", ProcessId);

   /* suppress the [delete] button for the main server process!! */
   if (ProcessId == HttpdProcessId)
      EndPageFunctionPtr = &WatchShowProcessEnd;
   else
      EndPageFunctionPtr = &WatchShowProcessDeleteEnd;

   rqptr->WatchShowProcessNextTaskFunction = NextTaskFunction;

   vecptr = FaoVector;
   *vecptr++ = ProcessIdString;
   *vecptr++ = ProcessIdString;
   *vecptr++ = ProcessIdString;
   *vecptr++ = ProcessIdString;
   *vecptr++ = ProcessIdString;

   status = sys$faol (&DclCommandFaoDsc, 0, &DclCommandDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   if (Debug) fprintf (stdout, "|%s|\n", DclCommand);

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);
   *vecptr++ = ProcessIdString;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   rqptr->NetWriteBufferedEscapeHtml = true;

   DclBegin (rqptr, EndPageFunctionPtr, DclCommand, NULL, NULL, NULL, NULL);
}

/*****************************************************************************/
/*
Called when the scripting subprocess is complete.  Output the last portion of
the report page and AST to wherever was the buffered end-of-report function.
*/

WatchShowProcessEnd (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (EndPageFaoDsc,
"</PRE></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchShowProcessEnd()\n");

   rqptr->NetWriteBufferedEscapeHtml = false;

   NetWriteBuffered (rqptr, rqptr->WatchShowProcessNextTaskFunction,
                     EndPageFaoDsc.dsc$a_pointer,
                     EndPageFaoDsc.dsc$w_length);
}

/*****************************************************************************/
/*
Called when the scripting subprocess is complete.  Output the last portion of
the report page and AST to wherever was the buffered end-of-report function.
*/

WatchShowProcessDeleteEnd (struct RequestStruct *rqptr)

{
   static $DESCRIPTOR (EndPageFaoDsc,
"</PRE></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
<P><FORM METHOD=GET ACTION=\"!AZ\">\n\
<INPUT TYPE=hidden NAME=pid VALUE=\"!AZ\">\n\
<INPUT TYPE=submit VALUE=\" Force Delete \">\n\
</FORM>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [32];
   char  Buffer [1024],
         ProcessIdString [32];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchShowProcessDeleteEnd()\n");

   rqptr->NetWriteBufferedEscapeHtml = false;

   if (!strsame (rqptr->QueryStringPtr, "pid=", 4))
   {
      WatchShowProcessEnd (rqptr);
      return;
   }
   strzcpy (ProcessIdString, rqptr->QueryStringPtr+4, sizeof(ProcessIdString));

   vecptr = FaoVector;
   *vecptr++ = INTERNAL_CONTROL_DELETE_PROCESS;
   *vecptr++ = ProcessIdString;

   status = sys$faol (&EndPageFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (rqptr->WatchShowProcessNextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, rqptr->WatchShowProcessNextTaskFunction,
                     Buffer, Length);
}

/*****************************************************************************/
/*
Just delete the process specified by 'ProcessIdString'.
*/

WatchDeleteProcess
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>Success 200</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>SUCCESS!!</H1>\n\
<P>Reported by server.\n\
<P>Server !AZ deleted process !8XL.\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;

   int  status;
   unsigned long  ProcessId;
   unsigned long  FaoVector [16];
   unsigned short  Length;
   char  Buffer [1024];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "WatchDeleteProcess() |%s|\n", rqptr->QueryStringPtr);

   if (strsame (rqptr->QueryStringPtr, "pid=", 4))
   {
      ProcessId = strtol (rqptr->QueryStringPtr+4, NULL, 16);
      if (Debug) fprintf (stdout, "ProcessID: %08.08X\n", ProcessId);
   }
   else
      ProcessId  = 0;

   if (ProcessId)
   {
      status = sys$delprc (&ProcessId, 0);
      if (Debug) fprintf (stdout, "sys$dlprc %%X%08.08X\n", status);
   }
   else
      status = SS$_BUGCHECK;

   if (VMSnok (status))
   {
      rqptr->ErrorTextPtr = "when deleting";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = ProcessId;

   sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, NextTaskFunction, Buffer, Length);
}

/*****************************************************************************/
/*
Provide a plain-text report page displaying some of the essential fields from
various data structures in an executing request.  Intended as a diagnosis and
development tool.
*/

WatchPeek
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *ConnectNumberString
)
{
#define WATCH_NULL_STRING "(null)"
#define WATCH_NULL(string) (string == NULL ? WATCH_NULL_STRING : string)

   static long  Subx2 = 2,
                EdivTen = 10;

   static unsigned long  DevNamItem = DVI$_DEVNAM;

   static $DESCRIPTOR (ServiceFaoDsc,
"!20%D  PEEK REPORT  !AZ\n\
!#*-\n\
!AZ\
\n\
!33<ServicePtr!> 0x!8XL\n\
!33<->ServerChannel!> !UL (!AZ)\n\
!33<->ServerHostPort!> |!AZ|\n\
!33<->ServerIpAddressString!> |!AZ|\n\
!33<->RequestSchemeNamePtr!> |!AZ|\n\
\n\
!33<ClientChannel!> !UL (!AZ)\n\
!33<ClientHostName!> |!AZ|\n\
!33<ClientIpAddressString!> |!AZ|\n\
!33<ClientPort!> !UL\n\
!33<NetReadRawAstFunctionPtr!> 0x!8XL\n\
!33<NetReadIOsb!> %X!8XL !UL\n\
!33<NetWriteRawAstFunctionPtr!> 0x!8XL\n\
!33<NetWriteIOsb!> %X!8XL !UL\n\
!33<SeSoLaPtr!> 0x!8XL\n");

   static $DESCRIPTOR (TimerFaoDsc,
"!33<BytesRawRx!> !UL\n\
!33<BytesRawTx!> !UL\n\
!33<KeepAliveCount!> !UL\n\
!33<KeepAliveRequest!> !UL\n\
!33<KeepAliveResponse!> !UL\n\
!33<KeepAliveTimeout!> !UL\n\
!33<TimerInputCount!> !UL\n\
!33<TimerKeepAliveCount!> !UL\n\
!33<TimerNoProgressBytesTx!> !UL\n\
!33<TimerNoProgressCount!> !UL\n\
!33<TimerNoProgressSeconds!> !UL\n\
!33<TimerOutputCount!> !UL\n\
!33<TimerTerminatedCount!> !UL\n\
\n\
!33<BinaryTime!> !%D (!UL.!#ZL seconds ago)\n\
!33<GmDateTime!> |!AZ|\n\
!33<RequestHeaderPtr!> |!#AZ|\n\
!33<RequestMethodName!> |!AZ|\n\
!33<ResourcePtr!> |!AZ|\n\
!33<HttpVersion!> !UL.!UL\n\
!33<HttpAcceptPtr!> |!AZ|\n\
!33<HttpAcceptCharsetPtr!> |!AZ|\n\
!33<HttpAcceptEncodingPtr!> |!AZ|\n\
!33<HttpAcceptLangPtr!> |!AZ|\n\
!33<HttpAuthorizationPtr!> |!AZ|\n\
!33<HttpContentTypePtr!> |!AZ|\n\
!33<HttpCookiePtr!> |!AZ|\n\
!33<HttpETagPtr!> |!AZ|\n\
!33<HttpForwardedPtr!> |!AZ|\n\
!33<HttpHostPtr!> |!AZ|\n\
!33<HttpIfModifiedSincePtr!> |!AZ|\n\
!33<HttpPragmaPtr!> |!AZ|\n\
!33<HttpProxyConnectionPtr!> |!AZ|\n\
!33<HttpRefererPtr!> |!AZ|\n\
!33<HttpUserAgentPtr!> |!AZ|\n\
!33<ContentLength!> !UL\n\
!33<ContentCount!> !UL\n\
!33<ContentBufferPtr!> 0x!XL\n\
\n\
!33<PathInfoPtr!> |!AZ|\n\
!33<QueryStringPtr!> |!AZ|\n\
!33<MappedPathPtr!> |!AZ|\n\
!33<RequestMappedFile!> |!AZ|\n\
!33<ScriptName!> |!AZ|\n\
!33<RequestMappedScript!> |!AZ|\n\
!33<RequestFileName!> |!AZ|\n\
!33<IsCgiPlusScript!> !UL\n\
\n\
!33<ResponseHeaderPtr!> |!#AZ|\n\
!33<ResponseHeaderSent!> !UL\n\
!33<ErrorReportPtr!> |!AZ|\n\
!33<LocationPtr!> |!AZ|\n\
\n\
!33<AuthorizationStatus!> %X!8XL\n\
!33<AuthRequestCan!> 0x!4XL (!AZ)\n\
!33<AuthUserCan!> 0x!4XL (!AZ)\n\
!33<AuthGroupCan!> 0x!4XL (!AZ)\n\
!33<AuthWorldCan!> 0x!4XL (!AZ)\n\
!33<AuthType!> |!AZ|\n\
!33<RemoteUser!> |!AZ|\n\
!33<RemoteUserPassword!> |!#**|\n\
!33<AuthUserDetailsPtr!> |!AZ|\n\
!33<AuthHttpsOnly!> !UL\n\
!33<AuthSysUafAuthenticated!> !UL\n\
!33<AuthVmsIdentifiersCount!> !UL\n\
!33<AuthVmsIdentifiersPtr!> 0x!XL\n\
!33<AuthVmsUserProfileLength!> !UL\n\
!33<AuthVmsUserProfilePtr!> 0x!XL\n\
!33<AuthRealmPtr!> |!AZ| (!AZ)\n\
!33<AuthRealmDescriptionPtr!> |!AZ|\n\
!33<AuthAgentParameterPtr!> |!AZ|\n\
!33<AuthGroupWritePtr!> |!AZ| (!AZ)\n\
!33<AuthGroupReadPtr!> |!AZ| (!AZ)\n\
!33<AuthGroupRestrictListPtr!> |!AZ|\n\
!33<AuthWorldRestrictListPtr!> |!AZ|\n\
\n\
!33<CachePtr!> 0x!XL\n\
!33<DclTaskPtr!> 0x!XL\n\
");

   static $DESCRIPTOR (DclTaskFaoDsc,
"!33<->CgiPlusInChannel!> !UL (!AZ)\n\
!33<->CgiPlusInIOsb!> %X!8XL !UL\n\
!33<->QueuedCgiPlusIn!> !UL\n\
!33<->HttpInputChannel!> !UL (!AZ)\n\
!33<->HttpInputIOsb!> %X!8XL !UL\n\
!33<->QueuedHttpInput!> !UL\n\
!33<->SysCommandChannel!> !UL (!AZ)\n\
!33<->SysCommandIOsb!> %X!8XL !UL\n\
!33<->QueuedSysCommand!> !UL\n\
!33<->QueuedSysCommandAllowed!> !UL\n\
!33<->SysOutputChannel!> !UL (!AZ)\n\
!33<->SysOutputIOsb!> %X!8XL !UL\n\
!33<->QueuedSysOutput!> !UL\n\
!33<->SubprocessPid!> !8XL\n\
!33<->SubprocessActivated!> !UL\n\
!33<->SubprocessResponded!> !UL\n\
!33<->SubprocessCompletionStatus!> %X!8XL\n\
!33<->MarkedForDelete!> !UL\n\
!33<->CgiPlusCalloutFunction!> 0x!8XL\n\
!33<->DclCommand!> |!AZ|\n\
!33<->ScriptName!> |!AZ|\n\
!33<->ScriptFileName!> |!AZ|\n\
\n");

   static $DESCRIPTOR (DECnetFaoDsc, "!33<DECnetTaskPtr!> 0x!8XL\n");

   static $DESCRIPTOR (DECnetTaskFaoDsc,
"!33<->DECnetChannel!> !UL\n\
!33<->DECnetConnectIOsb!> %X!8XL !UL\n\
!33<->DECnetReadIOsb!> %X!8XL !UL\n\
!33<->DECnetWriteIOsb!> %X!8XL !UL\n\
!33<->QueuedDECnetIO!> !UL\n\
!33<->ScriptResponded!> !UL\n\
!33<->CgiDialogState!> !UL\n\
!33<->OsuDialogState!> !UL\n\
\n");

   static $DESCRIPTOR (DescrFaoDsc,
"!33<DescrTaskPtr!> 0x!8XL\n\
!33<DirTaskPtr!> 0x!8XL\n\
!33<FileTaskPtr!> 0x!8XL\n\
!33<GraphTaskPtr!> 0x!8XL\n\
!33<HTAdminTaskPtr!> 0x!8XL\n\
!33<IsMapTaskPtr!> 0x!8XL\n\
!33<MenuTaskPtr!> 0x!8XL\n\
!33<ProxyTaskPtr!> 0x!8XL\n\
");

   static $DESCRIPTOR (ProxyTaskFaoDsc,
"!33<->ProxyChannel!> !UL (!AZ)\n\
!33<->RequestHostNamePort!> |!AZ|\n\
!33<->RequestSchemeName!> |!AZ|\n\
!33<->ProxyLookupIOsb!> %X!8XL !UL\n\
!33<->ProxyLookupRetryCount!> !UL\n\
!33<->ProxyConnectIOsb!> %X!8XL !UL\n\
!33<->ProxyReadIOsb!> %X!8XL !UL\n\
!33<->ProxyReadRawAstFunctionPtr!> 0x!8XL\n\
!33<->ProxyWriteIOsb!> %X!8XL !UL\n\
!33<->ProxyWriteRawAstFunctionPtr!> 0x!8XL\n\
!33<->BytesRawRx!> !UL\n\
!33<->BytesRawTx!> !UL\n\
!33<->RebuiltRequestPtr!> |!AZ|\n\
!33<->ResponseConsecutiveNewLineCount!> !UL\n\
!33<->ResponseHeaderLength!> !UL\n\
!33<->ResponseHeaderPtr!> |!#AZ|\n\
!33<->ResponseContentLength!> !SL\n\
!33<->ResponseBodyLength!> !UL\n\
!33<->ResponseBufferNetCount!> !UL\n\
!33<->ResponseBufferNetPtr!> 0x!8XL\n\
!33<->ResponseBufferCacheCount!> !UL\n\
!33<->ResponseBufferCachePtr!> 0x!8XL\n\
!33<->NotCacheable!> !UL\n\
!33<->ProxyCacheFileName!> |!AZ|\n\
!33<->ProxyCacheSuitable!> !UL\n\
!33<->ProxyCacheFileSizeInBytes!> !UL\n\
!33<->ProxyCacheLastModifiedHours!> !UL\n\
!33<->ProxyCacheReadBytes!> !UL\n\
!33<->ProxyCacheFileCdt!> !%D\n\
\n");

   static $DESCRIPTOR (PutFaoDsc,
"!33<PutTaskPtr!> 0x!8XL\n\
!33<SsiTaskPtr!> 0x!8XL\n\
!33<UpdTaskPtr!> 0x!8XL\n\
");

   register struct RequestStruct  *rqeptr;
   register struct ListEntryStruct  *leptr;

   int  status,
        ConnectNumber;
   unsigned short  Length;
   unsigned long  Remainder,
                  ResponseDuration;
   unsigned long  BinaryTime [2],
                  FaoVector [128],
                  ResultTime [2];
   unsigned long  *vecptr;
   char  Buffer [16384],
         ClientDevName [64],
         ProxyDevName [64],
         ServerDevName [64];
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (ClientDevNameDsc, ClientDevName);
   $DESCRIPTOR (ProxyDevNameDsc, ProxyDevName);
   $DESCRIPTOR (ServerDevNameDsc, ServerDevName);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchPeek() %s\n", ConnectNumberString);

   ConnectNumber = strtol (ConnectNumberString, NULL, 10);
   if (Debug) fprintf (stdout, "ConnectNumber: %d\n", ConnectNumber);

   if (ConnectNumber)
   {
      /* find this connection number in the current request list */
      for (leptr = RequestList.HeadPtr; leptr != NULL; leptr = leptr->NextPtr)
      {
         rqeptr = (struct RequestStruct*)leptr;
         if (Debug)
            fprintf (stdout, "ConnectNumber: %d\n", rqeptr->ConnectNumber);
         if (rqeptr->ConnectNumber == ConnectNumber) break;
      }
      if (leptr == NULL)
      {
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, ErrorWatchNumber, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
   }
   else
   {
      /* the current request */
      rqeptr = rqptr;
   }

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &rqeptr->ServicePtr->ServerChannel,
                   0, 0, &ServerDevNameDsc, &Length)))
      strzcpy (ServerDevName, SysGetMsg(status), sizeof(ServerDevName));
   else
      ServerDevName[Length] = '\0';

   if (VMSnok (status =
       lib$getdvi (&DevNamItem, &rqeptr->ClientChannel,
                   0, 0, &ClientDevNameDsc, &Length)))
      strzcpy (ClientDevName, SysGetMsg(status), sizeof(ClientDevName));
   else
      ClientDevName[Length] = '\0';

   sys$gettim (&BinaryTime);
   status = lib$subx (&BinaryTime, &rqeptr->BinaryTime, 
                      &ResultTime, &Subx2);
   /* convert to microseconds (longword becomes 71 minutes max) */
   if (VMSok (status))
      status = lib$ediv (&EdivTen, &ResultTime,
                         &ResponseDuration, &Remainder);
   if (Debug) fprintf (stdout, "elapsed: %%X%08.08X\n", status);
   if (VMSnok (status)) ResponseDuration = 0;

   rqptr->ResponsePreExpired = PRE_EXPIRE_WATCH;
   HTTP_HEADER_200_PLAIN (rqptr);

   vecptr = FaoVector;

   *vecptr++ = 0;
   *vecptr++ = ServerHostPort;
   *vecptr++ = 35 + strlen(ServerHostPort);
   *vecptr++ = WatchPackageInfo(rqptr);

   *vecptr++ = rqeptr->ServicePtr;
   *vecptr++ = rqeptr->ServicePtr->ServerChannel;
   *vecptr++ = ServerDevName+1;
   *vecptr++ = rqeptr->ServicePtr->ServerHostPort;
   *vecptr++ = rqeptr->ServicePtr->ServerIpAddressString;
   *vecptr++ = rqeptr->ServicePtr->RequestSchemeNamePtr;

   *vecptr++ = rqeptr->ClientChannel;
   *vecptr++ = ClientDevName+1;
   *vecptr++ = rqeptr->ClientHostName;
   *vecptr++ = rqeptr->ClientIpAddressString;
   *vecptr++ = rqeptr->ClientPort;
   *vecptr++ = rqeptr->NetReadRawAstFunctionPtr;
   *vecptr++ = rqeptr->NetReadIOsb.Status;
   *vecptr++ = rqeptr->NetReadIOsb.Count;
   *vecptr++ = rqeptr->NetWriteRawAstFunctionPtr;
   *vecptr++ = rqeptr->NetWriteIOsb.Status;
   *vecptr++ = rqeptr->NetWriteIOsb.Count;
   *vecptr++ = rqeptr->SeSoLaPtr;

   status = sys$faol (&ServiceFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   /* turn a leading date space into a leading zero */
   if (Buffer[0] == ' ' ) Buffer[0] = '0';

   NetWriteBuffered (rqptr, 0, Buffer, Length);

   SeSoLaWatchPeek (rqptr, rqeptr);

   vecptr = FaoVector;

   *vecptr++ = rqeptr->BytesRawRx;
   *vecptr++ = rqeptr->BytesRawTx;
   *vecptr++ = rqeptr->KeepAliveCount;
   *vecptr++ = rqeptr->KeepAliveRequest;
   *vecptr++ = rqeptr->KeepAliveResponse;
   *vecptr++ = rqeptr->KeepAliveTimeout;
   *vecptr++ = rqeptr->TimerInputCount;
   *vecptr++ = rqeptr->TimerKeepAliveCount;
   *vecptr++ = rqeptr->TimerNoProgressBytesTx;
   *vecptr++ = rqeptr->TimerNoProgressCount;
   *vecptr++ = rqeptr->TimerNoProgressSeconds;
   *vecptr++ = rqeptr->TimerOutputCount;
   *vecptr++ = rqeptr->TimerTerminatedCount;

   *vecptr++ = rqeptr->BinaryTime;
   *vecptr++ = ResponseDuration / USECS_IN_A_SECOND;
   *vecptr++ = DURATION_DECIMAL_PLACES;
   *vecptr++ = (ResponseDuration % USECS_IN_A_SECOND) / DURATION_UNITS_USECS;
   *vecptr++ = rqeptr->GmDateTime;
   if (rqeptr->RequestHeaderPtr == NULL)
      *vecptr++ = sizeof(WATCH_NULL_STRING)-1;
   else
      *vecptr++ = rqeptr->RequestHeaderLength;
   *vecptr++ = WATCH_NULL(rqeptr->RequestHeaderPtr);
   *vecptr++ = rqeptr->HttpMethodName;
   *vecptr++ = WATCH_NULL(rqeptr->ResourcePtr);
   *vecptr++ = rqeptr->HttpVersion / 10;
   *vecptr++ = rqeptr->HttpVersion % 10;
   *vecptr++ = WATCH_NULL(rqeptr->HttpAcceptPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpAcceptCharsetPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpAcceptEncodingPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpAcceptLangPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpAuthorizationPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpContentTypePtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpCookiePtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpETagPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpForwardedPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpHostPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpIfModifiedSincePtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpPragmaPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpProxyConnectionPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpRefererPtr);
   *vecptr++ = WATCH_NULL(rqeptr->HttpUserAgentPtr);
   *vecptr++ = rqeptr->ContentLength;
   *vecptr++ = rqeptr->ContentCount;
   *vecptr++ = rqeptr->ContentBufferPtr;

   *vecptr++ = WATCH_NULL(rqeptr->PathInfoPtr);
   *vecptr++ = WATCH_NULL(rqeptr->QueryStringPtr);
   *vecptr++ = WATCH_NULL(rqeptr->MappedPathPtr);
   *vecptr++ = rqeptr->RequestMappedFile;
   *vecptr++ = rqeptr->ScriptName;
   *vecptr++ = rqeptr->RequestMappedScript;
   *vecptr++ = rqeptr->RequestFileName;
   *vecptr++ = rqeptr->IsCgiPlusScript;

   if (rqeptr->ResponseHeaderPtr == NULL)
      *vecptr++ = sizeof(WATCH_NULL_STRING)-1;
   else
      *vecptr++ = rqeptr->ResponseHeaderLength;
   *vecptr++ = WATCH_NULL(rqeptr->ResponseHeaderPtr);
   *vecptr++ = rqeptr->ResponseHeaderSent;
   *vecptr++ = WATCH_NULL(rqeptr->ErrorReportPtr);
   *vecptr++ = WATCH_NULL(rqeptr->LocationPtr);

   *vecptr++ = rqeptr->AuthorizationStatus;
   *vecptr++ = rqeptr->AuthRequestCan;
   *vecptr++ = AuthCanString (rqeptr->AuthRequestCan, AUTH_CAN_FORMAT_LONG);
   *vecptr++ = rqeptr->AuthUserCan;
   *vecptr++ = AuthCanString (rqeptr->AuthUserCan, AUTH_CAN_FORMAT_LONG);
   *vecptr++ = rqeptr->AuthGroupCan;
   *vecptr++ = AuthCanString (rqeptr->AuthGroupCan, AUTH_CAN_FORMAT_LONG);
   *vecptr++ = rqeptr->AuthWorldCan;
   *vecptr++ = AuthCanString (rqeptr->AuthWorldCan, AUTH_CAN_FORMAT_LONG);
   *vecptr++ = rqeptr->AuthType;
   *vecptr++ = rqeptr->RemoteUser;
   *vecptr++ = strlen(rqeptr->RemoteUserPassword);
   *vecptr++ = WATCH_NULL(rqeptr->AuthUserDetailsPtr);
   *vecptr++ = rqeptr->AuthHttpsOnly;
   *vecptr++ = rqeptr->AuthSysUafAuthenticated;
   *vecptr++ = rqeptr->AuthVmsIdentifiersCount;
   *vecptr++ = rqeptr->AuthVmsIdentifiersPtr;
   *vecptr++ = rqeptr->AuthVmsUserProfileLength;
   *vecptr++ = rqeptr->AuthVmsUserProfilePtr;
   *vecptr++ = WATCH_NULL(rqeptr->AuthRealmPtr);
   *vecptr++ = AuthSourceString (rqeptr->AuthRealmPtr,
                                 rqeptr->AuthSourceRealm);
   *vecptr++ = WATCH_NULL(rqeptr->AuthRealmDescriptionPtr);
   *vecptr++ = WATCH_NULL(rqeptr->AuthAgentParameterPtr);
   *vecptr++ = WATCH_NULL(rqeptr->AuthGroupWritePtr);
   *vecptr++ = AuthSourceString (rqeptr->AuthGroupWritePtr,
                                 rqeptr->AuthSourceGroupWrite);
   *vecptr++ = WATCH_NULL(rqeptr->AuthGroupReadPtr);
   *vecptr++ = AuthSourceString (rqeptr->AuthGroupReadPtr,
                                 rqeptr->AuthSourceGroupRead);
   *vecptr++ = WATCH_NULL(rqeptr->AuthGroupRestrictListPtr);
   *vecptr++ = WATCH_NULL(rqeptr->AuthWorldRestrictListPtr);

   *vecptr++ = rqeptr->CachePtr;
   *vecptr++ = rqeptr->DclTaskPtr;

   status = sys$faol (&TimerFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (rqeptr->DclTaskPtr != NULL)
   {
      /************/
      /* DCL task */
      /************/

      vecptr = FaoVector;
      *vecptr++ = rqeptr->DclTaskPtr->CgiPlusInChannel;
      *vecptr++ = rqeptr->DclTaskPtr->CgiPlusInDevName+1;
      *vecptr++ = rqeptr->DclTaskPtr->CgiPlusInIOsb.Status;
      *vecptr++ = rqeptr->DclTaskPtr->CgiPlusInIOsb.Count;
      *vecptr++ = rqeptr->DclTaskPtr->QueuedCgiPlusIn;
      *vecptr++ = rqeptr->DclTaskPtr->HttpInputChannel;
      *vecptr++ = rqeptr->DclTaskPtr->HttpInputDevName+1;
      *vecptr++ = rqeptr->DclTaskPtr->HttpInputIOsb.Status;
      *vecptr++ = rqeptr->DclTaskPtr->HttpInputIOsb.Count;
      *vecptr++ = rqeptr->DclTaskPtr->QueuedHttpInput;
      *vecptr++ = rqeptr->DclTaskPtr->SysCommandChannel;
      *vecptr++ = rqeptr->DclTaskPtr->SysCommandDevName+1;
      *vecptr++ = rqeptr->DclTaskPtr->SysCommandIOsb.Status;
      *vecptr++ = rqeptr->DclTaskPtr->SysCommandIOsb.Count;
      *vecptr++ = rqeptr->DclTaskPtr->QueuedSysCommand;
      *vecptr++ = rqeptr->DclTaskPtr->QueuedSysCommandAllowed;
      *vecptr++ = rqeptr->DclTaskPtr->SysOutputChannel;
      *vecptr++ = rqeptr->DclTaskPtr->SysOutputDevName+1;
      *vecptr++ = rqeptr->DclTaskPtr->SysOutputIOsb.Status;
      *vecptr++ = rqeptr->DclTaskPtr->SysOutputIOsb.Count;
      *vecptr++ = rqeptr->DclTaskPtr->QueuedSysOutput;
      *vecptr++ = rqeptr->DclTaskPtr->SubprocessPid;
      *vecptr++ = rqeptr->DclTaskPtr->SubprocessActivated;
      *vecptr++ = rqeptr->DclTaskPtr->SubprocessResponded;
      *vecptr++ = rqeptr->DclTaskPtr->SubprocessCompletionStatus;
      *vecptr++ = rqeptr->DclTaskPtr->MarkedForDelete;
      *vecptr++ = rqeptr->DclTaskPtr->CgiPlusCalloutFunction;
      *vecptr++ = rqeptr->DclTaskPtr->DclCommand;
      *vecptr++ = rqeptr->DclTaskPtr->ScriptName;
      *vecptr++ = rqeptr->DclTaskPtr->ScriptFileName;

      status = sys$faol (&DclTaskFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;
   *vecptr++ = rqeptr->DECnetTaskPtr;

   status = sys$faol (&DECnetFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (rqeptr->DECnetTaskPtr != NULL)
   {
      /***************/
      /* DECnet task */
      /***************/

      vecptr = FaoVector;
      *vecptr++ = rqeptr->DECnetTaskPtr->DECnetChannel;
      *vecptr++ = rqeptr->DECnetTaskPtr->DECnetConnectIOsb.Status;
      *vecptr++ = rqeptr->DECnetTaskPtr->DECnetConnectIOsb.Count;
      *vecptr++ = rqeptr->DECnetTaskPtr->DECnetReadIOsb.Status;
      *vecptr++ = rqeptr->DECnetTaskPtr->DECnetReadIOsb.Count;
      *vecptr++ = rqeptr->DECnetTaskPtr->DECnetWriteIOsb.Status;
      *vecptr++ = rqeptr->DECnetTaskPtr->DECnetWriteIOsb.Count;
      *vecptr++ = rqeptr->DECnetTaskPtr->QueuedDECnetIO;
      *vecptr++ = rqeptr->DECnetTaskPtr->ScriptResponded;
      *vecptr++ = rqeptr->DECnetTaskPtr->CgiDialogState;
      *vecptr++ = rqeptr->DECnetTaskPtr->OsuDialogState;

      status = sys$faol (&DECnetTaskFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;
   *vecptr++ = rqeptr->DescrTaskPtr;
   *vecptr++ = rqeptr->DirTaskPtr;
   *vecptr++ = rqeptr->FileTaskPtr;
   *vecptr++ = rqeptr->GraphTaskPtr;
   *vecptr++ = rqeptr->HTAdminTaskPtr;
   *vecptr++ = rqeptr->IsMapTaskPtr;
   *vecptr++ = rqeptr->MenuTaskPtr;
   *vecptr++ = rqeptr->ProxyTaskPtr;

   status = sys$faol (&DescrFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (rqeptr->ProxyTaskPtr != NULL)
   {
      /**************/
      /* proxy task */
      /**************/

      if (VMSnok (status =
          lib$getdvi (&DevNamItem, &rqeptr->ProxyTaskPtr->ProxyChannel,
                      0, 0, &ProxyDevNameDsc, &Length)))
         strzcpy (ProxyDevName, SysGetMsg(status), sizeof(ProxyDevName));
      else
         ProxyDevName[Length] = '\0';

      vecptr = FaoVector;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyChannel;
      *vecptr++ = ProxyDevName+1;
      *vecptr++ = rqeptr->ProxyTaskPtr->RequestHostNamePort;
      *vecptr++ = rqeptr->ProxyTaskPtr->RequestSchemeName;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyLookupIOsb.Status;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyLookupIOsb.Count;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyLookupRetryCount;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyConnectIOsb.Status;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyConnectIOsb.Count;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyReadIOsb.Status;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyReadIOsb.Count;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyReadRawAstFunctionPtr;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyWriteIOsb.Status;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyWriteIOsb.Count;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyWriteRawAstFunctionPtr;
      *vecptr++ = rqeptr->ProxyTaskPtr->BytesRawRx;
      *vecptr++ = rqeptr->ProxyTaskPtr->BytesRawTx;
      *vecptr++ = WATCH_NULL(rqeptr->ProxyTaskPtr->RebuiltRequestPtr);
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseConsecutiveNewLineCount;
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseHeaderLength;
      if (rqeptr->ProxyTaskPtr->ResponseHeaderPtr == NULL)
      {
         *vecptr++ = sizeof(WATCH_NULL_STRING)-1;
         *vecptr++ = WATCH_NULL_STRING;
      }
      else
      if (rqeptr->ProxyTaskPtr->ResponseConsecutiveNewLineCount < 2)
      {
         *vecptr++ = rqeptr->ProxyTaskPtr->ResponseHeaderLength;
         *vecptr++ = WATCH_NULL(rqeptr->ProxyTaskPtr->ResponseHeaderPtr);
      }
      else
      {
         *vecptr++ = sizeof(WATCH_NULL_STRING)-1;
         *vecptr++ = WATCH_NULL_STRING;
      }
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseContentLength;
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseBodyLength;
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseBufferNetCount;
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseBufferNetPtr;
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseBufferCacheCount;
      *vecptr++ = rqeptr->ProxyTaskPtr->ResponseBufferCachePtr;
      *vecptr++ = rqeptr->ProxyTaskPtr->NotCacheable;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyCacheFileName;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyCacheSuitable;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyCacheFileSizeInBytes;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyCacheLastModifiedHours;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyCacheReadBytes;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyRequestNumber;
      *vecptr++ = rqeptr->ProxyTaskPtr->ProxyCacheFileCdt;

      status = sys$faol (&ProxyTaskFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;
   *vecptr++ = rqeptr->PutTaskPtr;
   *vecptr++ = rqeptr->SsiTaskPtr;
   *vecptr++ = rqeptr->UpdTaskPtr;

   status = sys$faol (&PutFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   NetWriteBuffered (rqptr, NextTaskFunction, NULL, 0);
}

/****************************************************************************/

